/**********************************************************************************

  DELTAGAMMAINC Fast and Accurate Evaluation of a Generalized Incomplete Gamma
  Function. Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com), Lionel
  Moisan (Lionel.Moisan AT parisdescartes.fr).

  This file is a part of the DELTAGAMMAINC software, dedicated to the
  computation of a generalized incomplete gammafunction. See the Companion paper
  for a complete description of the algorithm.

  ``Fast and accurate evaluation of a generalized incomplete gamma function''
  (Rémy Abergel, Lionel Moisan), preprint MAP5 nº2016-14.

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public License along with
  this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************************/

/* General definitions (generalized lower and upper incomplete gamma functions)
 * ----------------------------------------------------------------------------
 *
 * we define the generalized lower (gamma) and upper (Gamma) incomplete gamma
 * functions by
 *
 *   gamma(mu,p,x) = integral of s^{p-1} e^{-mu s} ds from s=0 to x
 *   Gamma(mu,p,x) = integral of s^{p-1} e^{-mu s} ds from s=x to +infinity
 *
 * where
 *
 *   + mu is a nonzero real number
 *   + p is a positive integer
 *   + x is a positive number that may takes the value + infinity (0 <= x <= infinity)
 *
 * Notice that:
 *
 *   + the value x=+infinity is allowed only when mu > 0
 *   + Gamma(mu,p,x) is defined only when mu >= 0
 *
 * Brief description of the following modules
 * ------------------------------------------
 *
 *
 *   + Gammaln approximates the log of the complete gamma function
 *
 *      Gamma(p) = integral of s^{p-1} e^{-s} ds from s=0 to +infinity.
 *
 *     using Pugh's method (a refinement of Lanczos algorithm).
 *
 *   + gammainc_ser evaluates the lower incomplete gamma function using a series
 *     expansion.
 *
 *   + gammainc_cfrac evaluates the lower incomplete gamma function using a
 *     continued fraction.
 *
 *   + gammainc_ibp evaluates the lower incomplete gamma function using a
 *     recursive integration by parts relation (WARNING: this function cannot be
 *     used when mu >= 0).
 *
 *   + Gammainc_cfrac evaluates the upper incomplete gamma function using a
 *     continued fraction.
 *
 * References
 * ----------
 *
 * + F. W. J. Olver, D. W. Lozier, R. F. Boisvert, and C. W. Clark
 *   (Eds.). 2010. NIST Handbook of Mathematical Functions. Cambridge University
 *   Press. (see online version at [[http://dlmf.nist.gov/]])
 *
 * + W. H. Press, S. A. Teukolsky, W. T. Vetterling, and
 *   B. P. Flannery. 1992. Numerical recipes in C: the art of scientific
 *   computing (2nd ed.).
 *
 * + G. R. Pugh, 2004. An analysis of then Lanczos Gamma approximation (phd
 *   thesis)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define ITMAX 10000 /* Maximum allowed number of iterations */
#define DPMIN 1.0e-300 /* Number near the smallest representable double-point number */
#define EPS 1.22e-16 /* Machine epsilon */

/* Gammaln: set (*gln) = log(Gamma(p)) for p > 0.
 * where Gamma(p) = integral of s^{p-1} e^{-s} ds from s=0 to +infinity.
 *
 * The evaluation is done using the Pugh's method (approximation with 11 terms),
 * which is a refinement of the Lanczos method (approximation with 6 terms).
 *
 *                                                                                    */
void Gammaln(double *gln, double p)
{
  double sum,l,z=p-1.;
  double d[11]={
    2.48574089138753565546e-5, 1.05142378581721974210e+0,
    -3.45687097222016235469e+0, 4.51227709466894823700e+0,
    -2.98285225323576655721e+0, 1.05639711577126713077e+0,
    -1.95428773191645869583e-1, 1.70970543404441224307e-2,
    -5.71926117404305781283e-4, 4.63399473359905636708e-6,
    -2.71994908488607703910e-9 };
  int k;

  for(sum=d[0],k=1; k<=10; k++) sum += d[k]/(z+(double)k);
  //*gln = log(2.) + 0.5*log(exp(1)/M_PI) + (z+0.5)*log((z+10.900511+0.5)/exp(1)) + log(sum);
  *gln = log(1.860382734205265717*sum) -(z+0.5) + (z+0.5L)*log(z+11.400511);
}


/* gammainc_ser: compute gamma(mu,p,x) using a series expansion
 *
 * usage:
 *
 *   + p: positive integer (p >= 1) casted in double
 *   + x: nonnegative number (0 <= x <= +infinity)
 *        notice that the value x = +infinity is allowed only when mu > 0
 *   + mu: nonzero real number
 *
 * this function computes mser and nser such as
 *
 *               gamma(a,x,m) = mser * exp(nser)
 *
 * where
 *
 * mser = sum_{k >= 0} ( (p-1)! / (k+p)! ) * (mu*x)^k
 * (this sum is truncated when the relative difference between
 * two consecutive terms is less than EPS)
 *
 * nser = - mu*x + p*log(x)
 *                                                                                    */
void gammainc_ser(double *mser, double *nser, double mu, double p, double x)
{

  double gln,d=p,del=1.0/p,sum=del,mux=mu*x;
  int n = 1;

  /* deal with special cases */
  if(x==0) { *mser = 0; *nser = -INFINITY; return; }
  if(isinf(x)){ *mser = 1; Gammaln(&gln,p); *nser = gln-p*log(mu); return; }

  /* evaluate the (truncated) serie */
  do { ++d; del *= mux/d; sum += del; n++; }
  while((fabs(del) >= fabs(sum)*EPS) && (n < ITMAX));
  *mser = sum; *nser = -mux + p*log(x);

  return;
}


/* gammainc_cfrac: compute gamma(mu,p,x) using a continued fraction
 *
 * usage :
 *
 *   + p: positive integer (p >= 1) casted in double
 *   + x: nonnegative number (0 <= x <= +infinity)
 *        notice that the value x = +infinity is allowed only when mu > 0
 *   + mu: nonzero real number
 *
 * this function computes mcfrac and ncfrac such as
 *
 *             gamma(mu,p,x) = mcfrac * exp(ncfrac)
 *
 * where
 *
 * mcfrac = a{1}/(b{1}+) a{2}/(b{2}+) a{3}/(b{3}+) ...
 * with a{1} = 1 and for k >= 1, a{2k} = -(p-1+k)*(mu*x),
 *      a{2k+1} = k*(mu*x) and b{k} = p-1+k.
 *
 * ncfrac = - mu*x + p*log(x)
 *                                                                                    */
void gammainc_cfrac(double *mcfrac, double *ncfrac, double mu, double p, double x)
{
  double gln,c,d,del,f,an,bn,mux=mu*x;
  int k,n;

  /* deal with special cases */
  if(x==0) { *mcfrac = 0; *ncfrac = -INFINITY; return; }
  if(isinf(x)){ *mcfrac = 1; Gammaln(&gln,p); *ncfrac = gln-p*log(mu); return; }

  /* evaluate the continued fraction using Modified Lentz's method. However, as
   * detailed in our paper, we perform manually the first pass (n=1), of the
   * initial Modified Lentz's method. */
  an=1.; bn=p; // set an = a{1}, bn = b{1} (b{1} is non-zero)
  f=an/bn; c=an/DPMIN; d=1./bn; n=2;
  do {
    k=n/2; an = ((n&1) ? k : -(p-1+k))*mux; bn++;
    d=an*d+bn; if (d == 0) d=DPMIN;
    c=bn+an/c; if (c == 0) c=DPMIN;
    d=1.0/d; del=d*c; f*=del; n++;
  }
  while((fabs(del-1.0) >= EPS) && (n < ITMAX));
  *mcfrac = f; *ncfrac = -mux+p*log(x);
}


/* gammainc_ibp: compute gamma(mu,p,x) using a recursive integration by parts
 *
 * usage :
 *
 *   + p: positive integer (p >= 1) casted in double
 *   + x: nonnegative number (0 <= x <= +infinity)
 *        notice that the value x = +infinity is allowed only when mu > 0
 *   + mu: negative real number (mu < 0)
 *
 * this function computes mibp and nibp such as
 *
 *             gamma(mu,p,x) = mibp * exp(nibp)
 *
 * where
 *
 * mibp = 1/(mu*x) * (exp(mu*x) * (p-1)! / (mu*x)^(p-1) - S)
 * with S = sum_{0 <= k <= p-1} (p-1)! * (mu*x)^{-k} / (p-1-k)!
 *
 * nibp = - mu*x + p*log(x)
 *                                                                                    */
void gammainc_ibp(double *mibp, double *nibp, double mu, double p, double x)
{
  double gln,t,tt,c,d,s,del;
  int l;
  char odd,stop;

  /* initialization */
  Gammaln(&gln,p); t = -mu*x; tt = 1./(t*t);
  odd = ((int)p)%2;

  /* evaluate the mantissa */
  c = 1./t; d = (p-1.); s = c*(t-d); l = 0;
  do {
    c *= d*(d-1)*tt;
    d -= 2.;
    del = c*(t-d);
    s += del; l++;
    stop = (fabs(del) < fabs(s)*EPS);
  }
  while((l<floor((p-2)/2)) && !stop);
  if(odd && !stop) { s += d*c/t; }
  *mibp = (((odd)?-1.:1.)*exp(-t+gln-(p-1)*log(t))+s)/t;

  /*
  else { // p is even
    c = 1./t; d = (p-1.); s = c*(t-d); l = 0;
    do { c *= d*(d-1)*tt; d -= 2.; del = c*(t-d); s += del; l++; }
    while((l<(p-2)/2) && (fabs(del) >= fabs(s)*EPS));
    *mibp = (exp(-t+gln-(p-1)*log(t))+s)/t;
  }
  */

  /* evaluate the exponent */
  *nibp = t+p*log(x);
}


/* Gammainc_cfrac: compute Gamma(a,x,m) using a continued fraction
 *
 * usage :
 *
 *   + p: positive integer (p >= 1) casted in double
 *   + x: nonnegative number (0 <= x <= +infinity)
 *   + mu: positive real number (mu > 0)
 *
 * this function computes Mcfrac and Ncfrac such as
 *
 *             Gamma(mu,p,x) = Mcfrac * exp(Ncfrac)
 *
 * where
 *
 * Mcfrac = a{1}/(b{1}+) a{2}/(b{2}+) a{3}/(b{3}+) ...
 * with a{1} = 1 and for n > 1, a{n} = -(n-1)*(n-p-1)
 * and for n >= 1 b{n} = mu*x + 2*n - p - 1.
 *
 * Ncfrac = - mu*x + p*log(x)
 *                                                                                    */
void Gammainc_cfrac(double *Mcfrac, double *Ncfrac, double mu, double p, double x)
{
  double gln,c,d,del,f,an,bn,mux=mu*x;
  int i,n;
  char t;

  /* deal with special cases */
  if(x==0) { *Mcfrac = 1; Gammaln(&gln,p); *Ncfrac = gln-p*log(mu); return; }
  if(isinf(x)) { *Mcfrac = 0; *Ncfrac = -INFINITY; return; }

  /* evaluate the continued fraction using Modified Lentz's method. However, as
   * detailed in our paper, we perform manually the first pass (n=1), of the
   * initial Modified Lentz's method. */
  an=1.,bn=mux+1.0-p; // set an = a{1}, bn = b{1}
  t = (bn != 0);
  if(t) { // b{1} is non-zero
    f=an/bn; c=an/DPMIN; d=1./bn; n=2;
  }
  else { // b{1}=0 but b{2} is non-zero, we compute Mcfrac = a{1}/f with f = a{2}/(b{2}+) a{3}/(b{3}+) ...
    an=-(1-p); bn=mux+3.0-p;
    f=an/bn; c=an/DPMIN; d=1./bn; n=3;
  }
  i = n-1;
  do {
    an = -i*(i-p); bn += 2.0;
    d=an*d+bn; if (d == 0) d=DPMIN;
    c=bn+an/c; if (c == 0) c=DPMIN;
    d=1.0/d; del=d*c; f*=del; i++; n++;
  }
  while((fabs(del-1.0) >= EPS) && (n < ITMAX));
  *Mcfrac = (t) ? f : 1./f; *Ncfrac = -mux+p*log(x);
}
