/**********************************************************************************

  DELTAGAMMAINC Fast and Accurate Evaluation of a Generalized Incomplete Gamma
  Function. Copyright (C) 2016 Remy Abergel (remy.abergel AT gmail.com), Lionel
  Moisan (Lionel.Moisan AT parisdescartes.fr).

  This file is a part of the DELTAGAMMAINC software, dedicated to the
  computation of a generalized incomplete gammafunction. See the Companion paper
  for a complete description of the algorithm.

  ``Fast and accurate evaluation of a generalized incomplete gamma function''
  (Rémy Abergel, Lionel Moisan), preprint MAP5 nº2016-14.

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.

  You should have received a copy of the GNU General Public License along with
  this program.  If not, see <http://www.gnu.org/licenses/>.

**********************************************************************************/

/*
 * Description: compute (rho,sigma) such as rho*exp(sigma) = I_{x,y}^{mu,p} where
 *
 *     I_{x,y}^{mu,p} = integral over [x,y] of s^(p-1) * exp(-mu*s) ds,
 *
 * using the extended double precision (i.e. using the 'long double' datatype)
 *
 * Compilation (with gcc):
 *
 *    gcc -O3 ldbl_gammainc.c ldbl_deltagammainc.c -lm -o ldbl_deltagammainc
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <string.h>
#include <math.h>

/* define flags for each possible computation method */
#define A_IS_GAMMA            (1<<1)
#define A_IS_UPPER_GINC_CFRAC (1<<2)
#define A_IS_LOWER_GINC_CFRAC (1<<3)
#define A_IS_LOWER_GINC_IBP   (1<<4)
#define B_IS_LOWER_GINC_CFRAC (1<<5)
#define B_IS_LOWER_GINC_IBP   (1<<6)
#define FLAG_TRAP  (1)                                             /* I is computed using a trapezoidal rule */
#define FLAG_DIFF1 (A_IS_GAMMA)                                    /* I=A-B, A=Gamma(p), B=(gammainc_cfrac(p,x)+Gammainc_cfrac(p,y)) */
#define FLAG_DIFF2 (A_IS_UPPER_GINC_CFRAC)                         /* I=A-B, A=Gammainc_cfrac(p,x), B=Gammainc_cfrac(p,x) */
#define FLAG_DIFF3 (A_IS_LOWER_GINC_CFRAC | B_IS_LOWER_GINC_CFRAC) /* I=A-B, A=gammainc_cfrac(p,y), B=gammainc_cfrac(p,x) */
#define FLAG_DIFF4 (A_IS_LOWER_GINC_IBP | B_IS_LOWER_GINC_IBP)     /* I=A-B, A=gammainc_ibp(p,y), B=gammainc_ibp(p,x) */
#define FLAG_DIFF5 (A_IS_LOWER_GINC_IBP | B_IS_LOWER_GINC_CFRAC)   /* I=A-B, A=gammainc_ibp(p,y), B=gammainc_cfrac(p,x) */

/* plim: compute plim(mu,x), the optimality frontier used to select which
 * difference A-B must be computed to approximate I_{x,y}^{mu,p} by I=A-B.
 *
 * This frontier is defined by
 *
 *               | mu * x             if   0 < mu*x
 *               |
 * plim(mu,x) = <     0               if -9 <= mu*x <= 0
 *               |
 *               | 5.*sqrt(mu*x)-5    otherwise
 *
 *
 *                                                                                       */
static double plim(long double mu, long double x)
{
  long double mux=mu*x;
  return (mu > 0) ? mux : ((mux >= -9.L) ? 0.L : 5.L*sqrtl(-mux)-5.L);
}

/* trapezoidal_rule: approximate I_{x,y}^{mu,p} using a trapezoidal rule (with one term).
 *
 * This procedure computes rho,sigma,m_err,n_err:
 *
 * (rho,sigma) are such as the approximated value of I_{x,y}^{mu,p} is
 * I=(*rho)*exp(*sigma)
 *
 * (m_err,n_err) are such as the quantity (*m_err)*exp(*n_err) is an upper bound
 * of the absolute error |I_{x,y}^{mu,p}-I|.
 *
 *                                                                                       */
static void trapezoidal_rule(long double *rho, long double *sig, long double *m_err, long double *n_err, long double mu, long double x, long double y, long double p)
{
  long double mux=mu*x,muy=mu*y,nmax=fmaxl(-mux,-muy),cx,cy,s0,nx,ny;

  /* compute (m_err,n_err) */
  if (p==1.L) {
    *m_err = mu*mu;
    *n_err = nmax;
  }
  else if (p==2) {
    cx = fabsl(mu*(mux-2.L));
    cy = fabsl(mu*(muy-2.L));
    *m_err = fmaxl(cx,cy);
    *n_err = nmax;
  }
  else {
    if(mu>0) {
      s0 = p-1.L;
      cx = fabsl((p-1.L)*(p-2.L)+mux*(mux-2.L*(p-1.L)));
      cy = fabsl((p-1.L)*(p-2.L)+muy*(muy-2.L*(p-1.L)));
      *m_err = ((x <= s0)&&(s0 <= y)) ? fmaxl(cx,fmax(cy,(p-1.L))) : fmaxl(cx,cy);
      *n_err = (p-3.L)*logl(y)+nmax;
    }
    else {
      *m_err = (p-1.L)*(p-2.L)-muy*(2.L*(p-1.L)-muy);
      *n_err = (p-3.L)*logl(y)+nmax;
    }
  }
  *m_err = (*m_err)/12.L;
  *n_err = (*n_err)+3.L*logl(y-x);

  /* compute (rho,sig) */
  cx = 0.5L*(y-x)/x; nx = p*logl(x)-mux;
  cy = 0.5L*(y-x)/y; ny = p*logl(y)-muy;
  *sig = fmaxl(nx,ny);
  *rho = cx*expl(nx-(*sig))+cy*expl(ny-(*sig));
}

/* deltagammainc: our algorithm for the approximation of I_{x,y}^{mu,p}.
 *
 * This procedure computes rho,sigma,method.
 *
 * (rho,sigma) are such as the approximated value of I_{x,y}^{mu,p} is
 * I=(*rho)*exp(*sigma)
 *
 * (*method) is a flag describing how I was computed (using a trapezoidal rule
 * or using some differences, in that case which difference).
 *
 */
static void deltagammainc(long double *rho, long double *sigma, char *method, long double x, long double y, long double mu, long double p)
{
  long double mA,mB,mx,My,nA,nB,nx,Ny,rho_trap,sig_trap,m_err,n_err,mux,muy,px,py,gln;

  /* initialization */
  mux = mu*x; muy = mu*y;
  px = plim(mu,x); py = plim(mu,y);
  *method = 0;

  /* kernel */
  if(x==y){ *rho = 0.L; *sigma = -INFINITY; }
  else {
    /* evaluate (m_err,n_err) and (rho_trap,sig_trap) such as
       rho_trap*exp(sig_trap) is the trapezoidal approximation of I_{x,y}^{mu,p}
       and m_err*exp(sig_err) is the an upper bound of the associated absolute
       error */
    trapezoidal_rule(&rho_trap,&sig_trap,&m_err,&n_err,mu,x,y,p);
    //
    /* now compute (mA,nA) and (mB,nB) such as I_{x,y}^{mu,p} can be
       approximated by the difference A-B, where A <= B, A = mA*exp(nA) and B =
       mB*exp(nB). An estimate of the absolute error associated to this
       approximation is 10.*A*LDBL_EPSILON. If this bound is higher than m1*exp(n1),
       approximate I_{x,y}^{mu,p} by (y-x)*y^{p}*exp(-muy). Otherwise,
       approximate I_{x,y}^{mu,p} by the difference A-B */
    if(mu<0) {
      if(p < py) { gammainc_ibp(&mA,&nA,mu,p,y); (*method) |= A_IS_LOWER_GINC_IBP; }
      else { gammainc_cfrac(&mA,&nA,mu,p,y); (*method) |= A_IS_LOWER_GINC_CFRAC; }
      if (p < px) { gammainc_ibp(&mB,&nB,mu,p,x); (*method) |=B_IS_LOWER_GINC_IBP; }
      else { gammainc_cfrac(&mB,&nB,mu,p,x); (*method) |= B_IS_LOWER_GINC_CFRAC; }
    }
    else {
      if(p<px) {
	Gammainc_cfrac(&mA,&nA,mu,p,x);
	Gammainc_cfrac(&mB,&nB,mu,p,y);
	(*method) = FLAG_DIFF2;
      }
      else if (py <= p) {
	gammainc_cfrac(&mA,&nA,mu,p,y);
	gammainc_cfrac(&mB,&nB,mu,p,x);
	(*method) = FLAG_DIFF3;
      }
      else {
	Gammaln(&gln,p); mA = 1; nA = gln-p*logl(mu);
	gammainc_cfrac(&mx,&nx,mu,p,x);
	Gammainc_cfrac(&My,&Ny,mu,p,y);
	nB = fmaxl(nx,Ny);
	if(isinf(nB)) nB = 0.L;
	mB = mx*expl(nx-nB) + My*expl(Ny-nB);
	(*method) = FLAG_DIFF1;
      }
    }
    /* compute (rho,sigma) such as rho*exp(sigma) = A-B */
    *rho = mA-mB*expl(nB-nA);
    *sigma = nA;
    /* compare the (estimated) absolute error between difference and trapezoidal rule */
    if((LDBL_EPSILON*(mA+mB*(1.L+fabsl(nB)+fabsl(nA))*expl(nB-nA)+(*rho)*fabsl(*sigma))/m_err)*expl((*sigma)-n_err) >= 1.L) {
      // select the first order approximation instead of the difference A-B.
      *rho = rho_trap;
      *sigma = sig_trap;
      (*method) = FLAG_TRAP;
    }
  }
}

static void display_usage()
{
  printf("Usage: ldbl_deltagammainc [--help] [--verbose] x y mu p\n\n");
  printf("   --help    : display help\n");
  printf("   --verbose : verbose mode (display how the integral was computed)\n");
  printf("   x         : (long double) a nonnegative number, possibly infinite (inf)\n");
  printf("   y         : (long double) a number, possibly infinite (inf), higher than x\n");
  printf("   mu        : (long double) a nonzero real number\n");
  printf("   p         : (long double) a real number >=1 (must be integer if mu<0)\n\n");
  printf("Description:\n\n");
  printf("   Compute (rho,sigma) such as I = rho * exp(sigma), where\n");
  printf("   I = integral over [x,y] of s^{p-1} * exp(-mu*s) ds.\n\n");
  printf("   This implementation uses the extended double precision, i.e. the 'long\n");
  printf("   double' datatype.\n\n");
}

int main(int argc, char **argv)
{
  char *x_value=NULL,*y_value=NULL,*mu_value=NULL,*p_value=NULL,vflag=NULL,method;
  long double rho,sigma,a,b,b0,l10,x,y,mu,p;
  int i,err;

  /* parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"--help") == 0) {
      display_usage();
      return EXIT_SUCCESS;
    }
    else if (strcmp(argv[i],"--verbose") == 0) {
      vflag = 1;
    }
    else if(NULL == x_value) {
      x_value = argv[i];
      err = sscanf(x_value,"%Lf",&x);
      if(err != 1) {
	printf("\nError: could not properly retrieve input argument 'x'.\n");
	display_usage();
	return EXIT_FAILURE;
      }
    }
    else if(NULL == y_value) {
      y_value = argv[i];
      err = sscanf(y_value,"%Lf",&y);
      if(err != 1) {
	printf("\nError: could not properly retrieve input argument 'y'.\n");
	display_usage();
	return EXIT_FAILURE;
      }
    }
    else if(NULL == mu_value) {
      mu_value = argv[i];
      err = sscanf(mu_value,"%Lf",&mu);
      if(err != 1) {
	printf("\nError: could not properly retrieve input argument 'mu'.\n");
	display_usage();
	return EXIT_FAILURE;
      }
    }
    else if(NULL == p_value) {
      p_value = argv[i];
      err = sscanf(p_value,"%Lf",&p);
      if(err != 1) {
	printf("\nError: could not properly retrieve input argument 'p'.\n");
	display_usage();
	return EXIT_FAILURE;
      }
    }
  }
  if ((NULL == x_value) || (NULL == y_value) || (NULL == mu_value) || (NULL == p_value)) {
    printf("\nError: invalid number of arguments.\n");
    display_usage();
    return EXIT_FAILURE;
  }

  /* chech consistency (can be removed to save time) */
  if(x<0) { printf("\nError: input 'x' must be nonnegative!\nType deltagammainc --help to display help.\n\n"); return EXIT_FAILURE; }
  if(y<0) { printf("\nError: input 'y' must be nonnegative!\nType deltagammainc --help to display help.\n\n"); return EXIT_FAILURE; }
  if(x>y) { printf("\nError: you must choose x <= y!\nType deltagammainc --help to display help.\n\n"); return EXIT_FAILURE; }
  if(isinf(y) && mu < 0) { printf("\nError: 'mu' < 0 is not allowed when y = +infinity!\nType deltagammainc --help to display help.\n\n"); return EXIT_FAILURE; }
  if(p < 1) { printf("\nError: you must choose input 'p' such as (p >= 1)!\nType deltagammainc --help to display help.\n\n"); return EXIT_FAILURE; }
  if((mu < 0) && (p != floor(p))) { printf("\nError: non integer values for input 'p' are not allowed when mu < 0!\nType deltagammainc --help to display help.\n\n"); return EXIT_FAILURE; }
  if (mu == 0) { printf("\nError: input 'mu' must be nonzero!\nType deltagammainc --help to display help.\n\n"); return EXIT_FAILURE; }

  /*** compute deltagammainc and printf result ***/
  printf("\nComputing I = integral over [x,y] of s^{p-1} * exp(-mu*s) ds,\n");
  printf("where (hexadecimal): x=%La, y=%La, mu=%La, p=%La,\n",x,y,mu,p);
  printf("(in decimal approx): x=%.34Lg, y=%.34Lg, mu=%.34Lg, p=%.34Lg.\n\n",x,y,mu,p);

  /***************** kernel *****************/
  deltagammainc(&rho,&sigma,&method,x,y,mu,p);
  /**************** end kernel ***************/

  /* deal with verbose mode (print additional information)*/
  if(vflag) {
    switch (method) {
    case FLAG_TRAP : // trapezoidal rule
      printf("Details: I will be computed using a first order trapezoidal rule.\n\n");
      break;
    case FLAG_DIFF1 : // I = Gamma(p) - (gammainc_cfrac(p,x) + Gammainc_cfrac(p,y))
      printf("Details: I will be computed as the difference I=A-B, where B=(B1+B2), and\n");
      printf("A  = Gamma(p)      = integral over [0,infinity) of s^{p-1} exp(-mu*s) ds,\n");
      printf("B1 = gammainc(p,x) = integral over [0,x] of s^{p-1} exp(-mu*s) ds,\n");
      printf("B2 = Gammainc(p,x) = integral over [y,infinity) of s^{p-1} exp(-mu*s) ds.\n");
      printf("A will be computed using a rational approximation (Pugh's method), B1 and \nB2 using some continued fractions.\n\n");
      break;
      case FLAG_DIFF2 : // I = Gammainc_cfrac(p,x) - Gammainc_cfrac(p,y)
	printf("Details: I will be computed as the difference I=A-B, where\n");
	printf("A = Gammainc(p,x) = integral over [x,infinity) of s^{p-1} exp(-mu*s) ds,\n");
	printf("B = Gammainc(p,y) = integral over [y,infinity) of s^{p-1} exp(-mu*s) ds,\n");
	printf("A and B will be computed using some continued fractions.\n\n");
	break;
      case FLAG_DIFF3 : // I = gammainc_cfrac(p,y) - gammainc_cfrac(p,x)
	printf("Details: I will be computed as the difference I=A-B, where\n");
	printf("A = gammainc(p,y) = integral over [0,y] of s^{p-1} exp(-mu*s) ds,\n");
	printf("B = gammainc(p,x) = integral over [0,x] of s^{p-1} exp(-mu*s) ds,\n");
	printf("A and B will be computed using some continued fractions.\n\n");
	break;
      case FLAG_DIFF4 : // I = gammainc_ibp(p,y) - gammainc_ibp(p,x)
	printf("Details: I will be computed as the difference I=A-B, where\n");
	printf("A = gammainc(p,y) = integral over [0,y] of s^{p-1} exp(-mu*s) ds,\n");
	printf("B = gammainc(p,x) = integral over [0,x] of s^{p-1} exp(-mu*s) ds,\n");
	printf("A and B will be computed using a recursive integration by parts.\n\n");
	break;
      case FLAG_DIFF5 : // I = gammainc_ibp(p,y) - gammainc_cfrac(p,x)
	printf("Details: I will be computed as the difference I=A-B, where\n");
	printf("A = gammainc(p,y) = integral over [0,y] of s^{p-1} exp(-mu*s) ds,\n");
	printf("B = gammainc(p,x) = integral over [0,x] of s^{p-1} exp(-mu*s) ds,\n");
	printf("A will be computed using a recursive integration by parts,\n");
	printf("B will be computed using a continued fraction.\n\n");
	break;
      }
  }

  /* manage display with decomposition I = a*10^(b) where a in [1,10) and b integer */
  if (rho==0 && isinf(sigma) && sigma < 0) printf("Output: I = %.30Le\n\n",0.L);
  else {
    a=rho*expl(sigma);
    //if(a>0. && !isinf(a)) printf("Output: I = %.17e\n\n",a);
    if(a>DBL_MIN && a<DBL_MAX) printf("Output: I = %.30Le\n\n",a);
    else {
      l10 = logl(10.L);
      b = sigma/l10;
      b0 = floorl(b+log10l(rho));
      a = rho*expl((b-b0)*l10);
      printf("Output: I = %.30Lfe%+03d\n\n",a,(int)b0);
    }
  }

  /* display the computed mantissa-exponent representation */
  printf("Representation with long double numbers: I = rho*exp(sigma), where\n");
  printf("rho   = %.30Le\n",rho);
  printf("sigma = %.30Le\n\n",sigma);

  return EXIT_SUCCESS;
}
